"""Utilities for managing the GraphQL schema."""

from __future__ import annotations

from collections import namedtuple
from dataclasses import fields, is_dataclass
from typing import Iterator, List, Type, Union, get_args, get_origin, get_type_hints

from apischema import deserialize as _deserialize
from apischema.metadata.keys import CONVERSION_METADATA, SKIP_METADATA
from apischema.type_names import get_type_name
from apischema.utils import to_camel_case
from gql.dsl import DSLField, DSLInlineFragment, DSLSchema

# For compatibility with < python 3.10.
NoneType = type(None)


def deserialize(*args, **kwargs):
    """Deserializes a GraphQL JSON blob.

    :meta private:
    """
    kwargs.setdefault("aliaser", to_camel_case)
    return _deserialize(*args, **kwargs)


_Field = namedtuple("_Field", ["name", "types"])


def _fields(cls, skip: list[str]) -> Iterator[_Field]:
    """Returns _Field objects for every field on the given dataclass.

    :meta private:
    """
    hints = get_type_hints(cls)
    for f in fields(cls):
        if f.name in skip:
            continue

        if (skip_md := f.metadata.get(SKIP_METADATA, None)) and (
            skip_md.serialization or skip_md.deserialization
        ):
            continue

        if conversion_md := f.metadata.get(CONVERSION_METADATA, None):
            yield _Field(f.name, [conversion_md.deserialization.source])
            continue

        field_type = hints[f.name]
        origin = get_origin(field_type)

        if origin == Union:
            # Drop None from Optional fields.
            field_types = set(get_args(field_type)) - {NoneType}

            # Actual unions just yield the union.
            if len(field_types) > 1:
                yield _Field(f.name, list(field_types))
                continue

            # If we have only one type left after dropping None, this could
            # still be a list. Perform the normal extraction routine.
            [field_type] = field_types
            origin = get_origin(field_type)

        if origin in (List, list):
            # Extract the contained type.
            # We assume all list types are uniform.
            [field_type] = get_args(field_type)

        yield _Field(f.name, [field_type])


def get_selectors(
    ds: DSLSchema, cls: Type, skip_fields: list[str] | None = None
) -> list[DSLField]:
    """Constructs GraphQL selectors for the given dataclass.

    :meta private:
    """
    ret = []
    skip_now, skip_later = parse_skip(skip_fields or [])
    for f in _fields(cls, skip_now):
        dsl_field = getattr(getattr(ds, get_type_name(cls).graphql), f.name)  # type: ignore[arg-type]
        if len(f.types) == 1:
            [f_type] = f.types
            if is_dataclass(f_type):
                f_skip = skip_later.get(f.name, [])
                ret.append(
                    getattr(dsl_field, "select")(*get_selectors(ds, f_type, f_skip))
                )
            else:
                ret.append(dsl_field)
        else:
            # This is a Union; we must pass an inline fragment for each type.
            sel_args = []
            for f_type in f.types:
                if not is_dataclass(f_type):
                    raise NotImplementedError
                sel_args.append(
                    DSLInlineFragment()
                    .on(getattr(ds, get_type_name(f_type).graphql))  # type: ignore[arg-type]
                    .select(*get_selectors(ds, f_type))
                )
            ret.append(getattr(dsl_field, "select")(*sel_args))
    return ret


def parse_skip(skip: list[str] | None = None) -> tuple[list[str], dict[str, list[str]]]:
    """Converts a flat list of skip fields into (skip_now, skip_later).

    skip_now is a list of fields in the current scope to skip.
    skip_later is a list of descendant fields to skip.

    :meta private:
    """
    now: list[str] = []
    later: dict[str, list[str]] = {}
    for item in skip or []:
        field, _, descendants = item.partition(".")
        if descendants:
            later.setdefault(field, []).append(descendants)
        else:
            now.append(field)
    return now, later



"""Authentication support for the Hydrawise v2 GraphQL API."""

from datetime import datetime, timedelta
from threading import Lock

import aiohttp

from .exceptions import NotAuthorizedError

CLIENT_ID = "hydrawise_app"
CLIENT_SECRET = "zn3CrjglwNV1"
TOKEN_URL = "https://app.hydrawise.com/api/v2/oauth/access-token"
DEFAULT_TIMEOUT = 60


class Auth:
    """Authentication support for the Hydrawise GraphQL API."""

    def __init__(self, username: str, password: str) -> None:
        """Initializer.

        :param username: The username to use for authenticating with the Hydrawise service.
        :param password: The password to use for authenticating with the Hydrawise service.
        """
        self.__username = username
        self.__password = password
        self._lock = Lock()
        self._token: str | None = None
        self._token_type: str | None = None
        self._token_expires: datetime | None = None
        self._refresh_token: str | None = None

    async def _fetch_token_locked(self, refresh=False):
        data = {
            "client_id": CLIENT_ID,
            "client_secret": CLIENT_SECRET,
        }
        if refresh:
            assert self._token is not None
            data["grant_type"] = "refresh_token"
            data["refresh_token"] = self._refresh_token
        else:
            data["grant_type"] = "password"
            data["scope"] = "all"
            data["username"] = self.__username
            data["password"] = self.__password
        async with aiohttp.ClientSession() as session:
            async with session.post(
                TOKEN_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data=data,
                timeout=DEFAULT_TIMEOUT,
            ) as resp:
                resp_json = await resp.json()
                if "error" in resp_json:
                    self._token_type = None
                    self._token = None
                    self._token_expires = None
                    raise NotAuthorizedError(resp_json["message"])
                self._token = resp_json["access_token"]
                self._refresh_token = resp_json["refresh_token"]
                self._token_type = resp_json["token_type"]
                self._token_expires = datetime.now() + timedelta(
                    seconds=resp_json["expires_in"]
                )

    async def check_token(self):
        """Checks a token and refreshes if necessary."""
        with self._lock:
            if self._token is None:
                await self._fetch_token_locked(refresh=False)
            elif self._token_expires - datetime.now() < timedelta(minutes=5):
                await self._fetch_token_locked(refresh=True)

    async def token(self) -> str:
        """Retrieves an authentication token for the current user.

        :rtype: string
        """
        await self.check_token()
        with self._lock:
            return f"{self._token_type} {self._token}"



from datetime import datetime, timedelta
import re
from unittest import mock

from aioresponses import aioresponses
from freezegun import freeze_time
from pytest import fixture

from pydrawise import legacy
from pydrawise.schema import Controller, Zone

API_KEY = "__api_key__"


@fixture
def customer_details():
    yield {
        "controller_id": 52496,
        "customer_id": 47076,
        "current_controller": "Home Controller",
        "controllers": [
            {
                "name": "Home Controller",
                "last_contact": 1693292420,
                "serial_number": "0310b36090",
                "controller_id": 52496,
                "status": "Unknown",
            },
            {
                "name": "Other Controller",
                "last_contact": 1693292420,
                "serial_number": "1310b36091",
                "controller_id": 63507,
                "status": "Unknown",
            },
        ],
    }


@fixture
def status_schedule():
    yield {
        "expanders": [],
        "master": 0,
        "master_post_timer": 0,
        "master_timer": 0,
        "message": "",
        "nextpoll": 60,
        "options": 1,
        "relays": [
            {
                "name": "Drips - House",
                "period": 259200,
                "relay": 1,
                "relay_id": 5965394,
                "run": 1800,
                "stop": 1,
                "time": 5400,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Drips - Fence",
                "period": 259200,
                "relay": 2,
                "relay_id": 5965395,
                "run": 1788,
                "stop": 1,
                "time": 1,
                "timestr": "Now",
                "type": 106,
            },
            {
                "name": "Rotary - Front",
                "period": 259200,
                "relay": 3,
                "relay_id": 5965396,
                "run": 1800,
                "stop": 1,
                "time": 1576800000,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Sprays - Side L",
                "period": 259200,
                "relay": 4,
                "relay_id": 5965397,
                "run": 180,
                "stop": 1,
                "time": 335997,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Rotary - Back N",
                "period": 259200,
                "relay": 5,
                "relay_id": 5965398,
                "run": 1800,
                "stop": 1,
                "time": 336177,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Rotary - Back C",
                "period": 259200,
                "relay": 6,
                "relay_id": 5965399,
                "run": 1800,
                "stop": 1,
                "time": 337977,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Rotary - Back F",
                "period": 259200,
                "relay": 7,
                "relay_id": 5965400,
                "run": 1200,
                "stop": 1,
                "time": 339777,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Sprays - Side R",
                "period": 259200,
                "relay": 8,
                "relay_id": 5965401,
                "run": 480,
                "stop": 1,
                "time": 340977,
                "timestr": "Sat",
                "type": 1,
            },
            {
                "name": "Sprays - Drivew",
                "period": 259200,
                "relay": 9,
                "relay_id": 5965402,
                "run": 900,
                "stop": 1,
                "time": 341457,
                "timestr": "Sat",
                "type": 1,
            },
        ],
        "sensors": [
            {
                "input": 0,
                "mode": 1,
                "offtimer": 0,
                "relays": [
                    {"id": 5965394},
                    {"id": 5965395},
                    {"id": 5965396},
                    {"id": 5965397},
                    {"id": 5965398},
                    {"id": 5965399},
                    {"id": 5965400},
                    {"id": 5965401},
                    {"id": 5965402},
                ],
                "timer": 0,
                "type": 1,
            }
        ],
        "simRelays": 1,
        "stupdate": 0,
        "time": 1693303803,
    }


@fixture
def success_status():
    yield {"message": "Successful message", "message_type": "info"}


@fixture
def mock_request(customer_details, status_schedule):
    with mock.patch("requests.get") as req:
        controller_info_resp = mock.Mock(return_code=200)
        controller_info_resp.json.return_value = customer_details
        controller_status_resp = mock.Mock(return_code=200)
        controller_status_resp.json.return_value = status_schedule
        req.side_effect = [controller_info_resp, controller_status_resp]
        yield req


class TestLegacyHydrawiseAsync:
    """Test the LegacyHydrawiseAsync class."""

    async def test_get_user(
        self, customer_details: dict, status_schedule: dict
    ) -> None:
        """Test the get_user method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with freeze_time("2023-01-01 01:00:00") as t:
            with aioresponses() as m:
                m.get(
                    f"https://api.hydrawise.com/api/v1/customerdetails.php?api_key={API_KEY}",
                    status=200,
                    payload=customer_details,
                )
                m.get(
                    f"https://api.hydrawise.com/api/v1/statusschedule.php?api_key={API_KEY}&controller_id=52496",
                    status=200,
                    payload=status_schedule,
                )
                m.get(
                    f"https://api.hydrawise.com/api/v1/statusschedule.php?api_key={API_KEY}&controller_id=63507",
                    status=200,
                    payload=status_schedule,
                )
                user = await client.get_user()
                assert user.customer_id == 47076
                assert [c.id for c in user.controllers] == [52496, 63507]
                want_zones = [
                    5965394,
                    5965395,
                    5965396,
                    5965397,
                    5965398,
                    5965399,
                    5965400,
                    5965401,
                    5965402,
                ]
                assert [z.id for z in user.controllers[0].zones] == want_zones
                assert [z.id for z in user.controllers[1].zones] == want_zones

    async def test_get_controllers(
        self, customer_details: dict, status_schedule: dict
    ) -> None:
        """Test the get_controllers method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with freeze_time("2023-01-01 01:00:00") as t:
            with aioresponses() as m:
                m.get(
                    f"https://api.hydrawise.com/api/v1/customerdetails.php?api_key={API_KEY}&type=controllers",
                    status=200,
                    payload=customer_details,
                )
                m.get(
                    f"https://api.hydrawise.com/api/v1/statusschedule.php?api_key={API_KEY}&controller_id=52496",
                    status=200,
                    payload=status_schedule,
                )
                m.get(
                    f"https://api.hydrawise.com/api/v1/statusschedule.php?api_key={API_KEY}&controller_id=63507",
                    status=200,
                    payload=status_schedule,
                )
                controllers = await client.get_controllers()
                assert [c.id for c in controllers] == [52496, 63507]
                assert controllers[0].name == "Home Controller"
                assert controllers[1].name == "Other Controller"
                assert controllers[0].hardware.serial_number == "0310b36090"
                assert controllers[1].hardware.serial_number == "1310b36091"
                want_last_contact_time = datetime(2023, 8, 29, 7, 0, 20)
                assert controllers[0].last_contact_time == want_last_contact_time
                assert controllers[1].last_contact_time == want_last_contact_time
                want_zones = [
                    5965394,
                    5965395,
                    5965396,
                    5965397,
                    5965398,
                    5965399,
                    5965400,
                    5965401,
                    5965402,
                ]
                assert [z.id for z in controllers[0].zones] == want_zones
                assert [z.id for z in controllers[1].zones] == want_zones

    async def test_get_zones(self, status_schedule: dict) -> None:
        """Test the get_zones method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with freeze_time("2023-01-01 01:00:00") as t:
            with aioresponses() as m:
                m.get(
                    f"https://api.hydrawise.com/api/v1/statusschedule.php?api_key={API_KEY}&controller_id=12345",
                    status=200,
                    payload=status_schedule,
                )
                zones = await client.get_zones(Controller(id=12345))
                assert [z.id for z in zones] == [
                    5965394,
                    5965395,
                    5965396,
                    5965397,
                    5965398,
                    5965399,
                    5965400,
                    5965401,
                    5965402,
                ]
                assert zones[0].name == "Drips - House"
                assert zones[0].number == 1
                assert zones[0].scheduled_runs.current_run is None
                next_run = zones[0].scheduled_runs.next_run
                assert next_run.start_time == datetime(2023, 1, 1, 2, 30)
                assert next_run.normal_duration == timedelta(seconds=1800)
                assert next_run.duration == timedelta(seconds=1800)

                assert zones[1].name == "Drips - Fence"
                assert zones[1].number == 2
                current_run = zones[1].scheduled_runs.current_run
                assert current_run.start_time == datetime(2023, 1, 1, 1, 0, 0)
                assert current_run.end_time == datetime(2023, 1, 1, 1, 0, 0)
                assert current_run.normal_duration == timedelta(minutes=0)
                assert current_run.duration == timedelta(minutes=0)
                assert current_run.remaining_time == timedelta(seconds=1788)
                assert zones[1].scheduled_runs.next_run is None

                assert zones[2].name == "Rotary - Front"
                assert zones[2].number == 3
                assert zones[2].scheduled_runs.current_run is None
                assert zones[2].scheduled_runs.next_run is None
                assert zones[2].status.suspended_until == datetime.max

    async def test_start_zone(self, success_status: dict) -> None:
        """Test the start_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            zone = mock.create_autospec(Zone)
            zone.id = 12345
            await client.start_zone(zone)
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "run",
                    "relay_id": 12345,
                    "period_id": 999,
                },
                timeout=10,
            )

    async def test_stop_zone(self, success_status: dict) -> None:
        """Test the stop_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            zone = mock.create_autospec(Zone)
            zone.id = 12345
            await client.stop_zone(zone)
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={"api_key": API_KEY, "action": "stop", "relay_id": 12345},
                timeout=10,
            )

    async def test_start_all_zones(self, success_status: dict) -> None:
        """Test the start_all_zones method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            await client.start_all_zones(Controller(id=1111))
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "runall",
                    "period_id": 999,
                    "controller_id": 1111,
                },
                timeout=10,
            )

    async def test_stop_all_zones(self, success_status: dict) -> None:
        """Test the stop_all_zones method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            await client.stop_all_zones(Controller(id=1111))
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={"api_key": API_KEY, "action": "stopall", "controller_id": 1111},
                timeout=10,
            )

    async def test_suspend_zone(self, success_status: dict) -> None:
        """Test the suspend_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            zone = mock.create_autospec(Zone)
            zone.id = 12345
            await client.suspend_zone(zone, datetime(2023, 1, 2, 1, 0))
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspend",
                    "relay_id": 12345,
                    "period_id": 999,
                    "custom": 1672621200,
                },
                timeout=10,
            )

    async def test_resume_zone(self, success_status: dict) -> None:
        """Test the resume_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            zone = mock.create_autospec(Zone)
            zone.id = 12345
            await client.resume_zone(zone)
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspend",
                    "relay_id": 12345,
                    "period_id": 0,
                },
                timeout=10,
            )

    async def test_suspend_all_zones(self, success_status: dict) -> None:
        """Test the suspend_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            await client.suspend_all_zones(
                Controller(id=1111), datetime(2023, 1, 2, 1, 0)
            )
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspendall",
                    "period_id": 999,
                    "custom": 1672621200,
                    "controller_id": 1111,
                },
                timeout=10,
            )

    async def test_resume_all_zones(self, success_status: dict) -> None:
        """Test the suspend_zone method."""
        client = legacy.LegacyHydrawiseAsync(API_KEY)
        with aioresponses() as m:
            m.get(
                re.compile("https://api.hydrawise.com/api/v1/setzone.php"),
                status=200,
                payload=success_status,
            )
            await client.resume_all_zones(Controller(id=1111))
            m.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspendall",
                    "period_id": 0,
                    "controller_id": 1111,
                },
                timeout=10,
            )


class TestLegacyHydrawise:
    def test_update(self, mock_request, customer_details, status_schedule):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.assert_has_calls(
            [
                mock.call(
                    "https://api.hydrawise.com/api/v1/customerdetails.php",
                    params={"api_key": API_KEY, "type": "controllers"},
                    timeout=10,
                ),
                mock.call(
                    "https://api.hydrawise.com/api/v1/statusschedule.php",
                    params={"api_key": API_KEY},
                    timeout=10,
                ),
            ]
        )
        assert client.controller_info == customer_details
        assert client.controller_status == status_schedule

    def test_attributes(self, mock_request, customer_details, status_schedule):
        client = legacy.LegacyHydrawise(API_KEY)
        assert client.current_controller == customer_details["controllers"][0]
        assert client.status == "Unknown"
        assert client.controller_id == 52496
        assert client.customer_id == 47076
        assert client.num_relays == 9
        assert client.relays == status_schedule["relays"]
        assert list(client.relays_by_id.keys()) == [
            5965394,
            5965395,
            5965396,
            5965397,
            5965398,
            5965399,
            5965400,
            5965401,
            5965402,
        ]
        assert list(client.relays_by_zone_number.keys()) == [1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert client.name == "Home Controller"
        assert client.sensors == status_schedule["sensors"]
        assert client.running is None

    @mock.patch("requests.get")
    def test_attributes_not_initialized(self, mock_request):
        mock_request.side_effect = NotImplementedError
        client = legacy.LegacyHydrawise(API_KEY, load_on_init=False)
        assert client.controller_info == {}
        assert client.controller_status == {}
        assert client.current_controller == {}
        assert client.status is None
        assert client.controller_id is None
        assert client.customer_id is None
        assert client.num_relays == 0
        assert client.relays == []
        assert client.relays_by_id == {}
        assert client.relays_by_zone_number == {}
        assert client.name is None
        assert client.sensors == []
        assert client.running is None

    def test_suspend_zone(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.suspend_zone(1, 1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspend",
                    "custom": 1672617600,
                    "period_id": 999,
                    "relay_id": 5965394,
                },
                timeout=10,
            )

    def test_suspend_zone_unsuspend(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.suspend_zone(0, 1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspend",
                    "period_id": 0,
                    "relay_id": 5965394,
                },
                timeout=10,
            )

    def test_suspend_zone_all(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.suspend_zone(1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "suspendall",
                    "custom": 1672617600,
                    "period_id": 999,
                },
                timeout=10,
            )

    def test_run_zone(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.run_zone(1, 1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "run",
                    "custom": 60,
                    "period_id": 999,
                    "relay_id": 5965394,
                },
                timeout=10,
            )

    def test_run_zone_all(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.run_zone(1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "runall",
                    "custom": 60,
                    "period_id": 999,
                },
                timeout=10,
            )

    def test_run_zone_stop(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.run_zone(0, 1) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "stop",
                    "period_id": 0,
                    "relay_id": 5965394,
                },
                timeout=10,
            )

    def test_run_zone_stop_all(self, mock_request, success_status):
        client = legacy.LegacyHydrawise(API_KEY)
        mock_request.reset_mock(return_value=True, side_effect=True)

        mock_request.return_value.status_code = 200
        mock_request.return_value.json.return_value = success_status

        with freeze_time("2023-01-01 00:00:00") as t:
            assert client.run_zone(0) == success_status
            mock_request.assert_called_once_with(
                "https://api.hydrawise.com/api/v1/setzone.php",
                params={
                    "api_key": API_KEY,
                    "action": "stopall",
                    "period_id": 0,
                },
                timeout=10,
            )



"""Base class for the Hydrawise client API."""

from abc import ABC, abstractmethod
from datetime import datetime

from .schema import Controller, User, Zone, ZoneSuspension


class HydrawiseBase(ABC):
    """Base class for Hydrawise client APIs."""

    @abstractmethod
    async def get_user(self, fetch_zones: bool = True) -> User:
        """Retrieves the currently authenticated user.

        :param fetch_zones: When True, also fetch zones.
        :rtype: User
        """

    @abstractmethod
    async def get_controllers(self) -> list[Controller]:
        """Retrieves all controllers associated with the currently authenticated user.

        :rtype: list[Controller]
        """

    @abstractmethod
    async def get_controller(self, controller_id: int) -> Controller:
        """Retrieves a single controller by its unique identifier.

        :param controller_id: Unique identifier for the controller to retrieve.
        :rtype: Controller
        """

    @abstractmethod
    async def get_zones(self, controller: Controller) -> list[Zone]:
        """Retrieves zones associated with the given controller.

        :param controller: Controller whose zones to fetch.
        :rtype: list[Zone]
        """

    @abstractmethod
    async def get_zone(self, zone_id: int) -> Zone:
        """Retrieves a zone by its unique identifier.

        :param zone_id: The zone's unique identifier.
        :rtype: Zone
        """

    @abstractmethod
    async def start_zone(
        self,
        zone: Zone,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts a zone's run cycle.

        :param zone: The zone to start.
        :param mark_run_as_scheduled: Whether to mark the zone as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zone. If not
            specified (or zero), will run for its default configured time.
        """

    @abstractmethod
    async def stop_zone(self, zone: Zone) -> None:
        """Stops a zone.

        :param zone: The zone to stop.
        """

    @abstractmethod
    async def start_all_zones(
        self,
        controller: Controller,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts all zones attached to a controller.

        :param controller: The controller whose zones to start.
        :param mark_run_as_scheduled: Whether to mark the zones as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zones. If not
            specified (or zero), will run for each zone's default configured time.
        """

    @abstractmethod
    async def stop_all_zones(self, controller: Controller) -> None:
        """Stops all zones attached to a controller.

        :param controller: The controller whose zones to stop.
        """

    @abstractmethod
    async def suspend_zone(self, zone: Zone, until: datetime) -> None:
        """Suspends a zone's schedule.

        :param zone: The zone to suspend.
        :param until: When the suspension should end.
        """

    @abstractmethod
    async def resume_zone(self, zone: Zone) -> None:
        """Resumes a zone's schedule.

        :param zone: The zone whose schedule to resume.
        """

    @abstractmethod
    async def suspend_all_zones(self, controller: Controller, until: datetime) -> None:
        """Suspends the schedule of all zones attached to a given controller.

        :param controller: The controller whose zones to suspend.
        :param until: When the suspension should end.
        """

    @abstractmethod
    async def resume_all_zones(self, controller: Controller) -> None:
        """Resumes the schedule of all zones attached to the given controller.

        :param controller: The controller whose zones to resume.
        """

    @abstractmethod
    async def delete_zone_suspension(self, suspension: ZoneSuspension) -> None:
        """Removes a specific zone suspension.

        Useful when there are multiple suspensions for a zone in effect.

        :param suspension: The suspension to delete.
        """



from datetime import timedelta

from aioresponses import aioresponses
from freezegun import freeze_time
from pytest import fixture

from pydrawise import auth


@fixture
def token_payload():
    return {
        "access_token": "__access-token__",
        "refresh_token": "__refresh-token__",
        "token_type": "bearer",
        "expires_in": 360,  # 6 minutes
    }


@fixture
def mock_token_fetch(token_payload):
    with aioresponses() as m:
        m.post(
            auth.TOKEN_URL,
            status=200,
            payload=token_payload,
        )
        yield m


async def test_check_token_fetch_and_refresh(token_payload):
    a = auth.Auth("__username__", "__password__")
    with freeze_time("2023-01-01 01:00:00") as t:
        # Initial fetch
        with aioresponses() as m:
            m.post(auth.TOKEN_URL, status=200, payload=token_payload)
            await a.check_token()
            m.assert_called_once_with(
                auth.TOKEN_URL,
                method="POST",
                data={
                    "client_id": auth.CLIENT_ID,
                    "client_secret": auth.CLIENT_SECRET,
                    "grant_type": "password",
                    "scope": "all",
                    "username": "__username__",
                    "password": "__password__",
                },
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                timeout=auth.DEFAULT_TIMEOUT,
            )

        # Token refresh
        t.tick(delta=timedelta(minutes=2))
        with aioresponses() as m:
            m.post(auth.TOKEN_URL, status=200, payload=token_payload)
            await a.check_token()
            m.assert_called_once_with(
                auth.TOKEN_URL,
                method="POST",
                data={
                    "client_id": auth.CLIENT_ID,
                    "client_secret": auth.CLIENT_SECRET,
                    "grant_type": "refresh_token",
                    "refresh_token": "__refresh-token__",
                },
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                timeout=auth.DEFAULT_TIMEOUT,
            )


async def test_token(mock_token_fetch):
    a = auth.Auth("__username__", "__password__")
    token = await a.token()
    assert token == "bearer __access-token__"



"""GraphQL API schema for pydrawise."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, time, timedelta, timezone
from enum import Enum, auto
from typing import Optional, Union

from apischema import type_name
from apischema.conversions import Conversion
from apischema.metadata import conversion, fall_back_on_default

# The names in this file are from the GraphQL schema and don't always adhere to
# the naming scheme that pylint expects.
# pylint: disable=invalid-name


def _optional_field(*args, **kwargs):
    if "metadata" in kwargs:
        kwargs["metadata"] |= fall_back_on_default
    else:
        kwargs["metadata"] = fall_back_on_default
    return field(*args, **kwargs)


def default_datetime() -> datetime:
    """Default datetime factory for fields in this module.

    Abstracted so it can be mocked out.

    :meta private:
    """
    return datetime.now()


def _duration_conversion(unit: str) -> conversion:
    assert unit in (
        "days",
        "seconds",
        "microseconds",
        "milliseconds",
        "minutes",
        "hours",
        "weeks",
    )
    return conversion(
        Conversion(lambda d: timedelta(**{unit: d}), source=int, target=timedelta),
        Conversion(lambda d: getattr(d, unit), source=timedelta, target=int),
    )


def _timestamp_conversion() -> conversion:
    return conversion(
        Conversion(datetime.fromtimestamp, source=int, target=datetime),
        Conversion(datetime.timestamp, source=datetime, target=int),
    )


def _time_conversion() -> conversion:
    return conversion(
        Conversion(
            lambda s: datetime.strptime(s, "%H:%M").time(), source=str, target=time
        ),
        Conversion(lambda t: t.strftime("%H:%M"), source=time, target=str),
    )


def _list_conversion(element_conversion) -> conversion:
    return conversion(
        Conversion(
            lambda l: list(map(element_conversion.deserialization.converter, l)),
            source=list,
            target=list,
        ),
        Conversion(
            lambda l: list(map(element_conversion.serialization.converter, l)),
            source=list,
            target=list,
        ),
    )


class _AutoEnum(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        """Determines the value for an auto() call."""
        return name


class StatusCodeEnum(_AutoEnum):
    """Response status codes."""

    OK = auto()
    WARNING = auto()
    ERROR = auto()


@dataclass
class StatusCodeAndSummary:
    """A response status code and a human-readable summary."""

    status: StatusCodeEnum = StatusCodeEnum.OK
    summary: str = ""


@dataclass
class LocalizedValueType:
    """A localized value."""

    value: float = _optional_field(default=0.0)
    unit: str = _optional_field(default="")


@dataclass
class SelectedOption:
    """A generic option."""

    value: int = 0
    label: str = _optional_field(default="")


@dataclass
class DateTime:
    """A date & time.

    This is only used for serialization and deserialization.
    """

    value: str = ""
    timestamp: int = 0

    @staticmethod
    def from_json(dt: DateTime) -> datetime:
        """Converts a DateTime to a native python type."""
        return datetime.fromtimestamp(dt.timestamp)

    @staticmethod
    def to_json(dt: datetime) -> DateTime:
        """Converts a native datetime to a DateTime GraphQL type."""
        local = dt
        if local.tzinfo is None:
            # Make sure we have a timezone set so strftime outputs a valid string.
            local = local.replace(tzinfo=datetime.now(timezone.utc).astimezone().tzinfo)
        return DateTime(
            value=local.strftime("%a, %d %b %y %H:%I:%S %z"),
            timestamp=int(dt.timestamp()),
        )

    @staticmethod
    def conversion() -> conversion:
        """Returns a GraphQL conversion for a DateTime."""
        return conversion(
            Conversion(DateTime.from_json, source=DateTime, target=datetime),
            Conversion(DateTime.to_json, source=datetime, target=DateTime),
        )


@type_name("Zone")
@dataclass
class BaseZone:
    """Basic zone information."""

    id: int = 0
    number: SelectedOption = field(default_factory=SelectedOption)
    name: str = ""


@dataclass
class CycleAndSoakSettings:
    """Cycle and soak durations."""

    cycle_duration: timedelta = field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )
    soak_duration: timedelta = field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )


@dataclass
class RunTimeGroup:
    """The runtime of a watering program group."""

    id: int = 0
    name: str = _optional_field(default="")
    duration: timedelta = field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )


@dataclass
class WateringPeriodicity:
    """Watering frequency description (e.g., "Every Program Start Time")."""

    value: int = _optional_field(default=0)
    label: str = _optional_field(default="")


@dataclass
class ProgramWateringFrequency:
    """Watering frequency information."""

    label: str = ""
    period: WateringPeriodicity = field(default_factory=WateringPeriodicity)
    description: str = ""


@dataclass
@type_name("StandardProgram")
class StandardProgramRef:
    """Super small base class to reference a watering program without having
    to pull in all the voluminous sub-fields."""

    id: int = 0
    name: str = ""


@dataclass
@type_name("AdvancedProgram")
class AdvancedProgramRef:
    """Super small base class to reference a watering program without having
    to pull in all the voluminous sub-fields."""

    id: int = 0
    name: str = ""


@dataclass
class Program:
    """Base class for a watering program."""

    id: int = 0
    name: str = ""

    scheduling_method: SelectedOption = field(default_factory=SelectedOption)
    monthly_watering_adjustments: list[int] = field(default_factory=list)
    applies_to_zones: list[BaseZone] = field(default_factory=list)


@dataclass
class AdvancedProgram(Program):
    """An advanced watering program."""

    zone_specific: bool = False
    advanced_program_id: int = 0
    watering_frequency: ProgramWateringFrequency = _optional_field(
        default_factory=ProgramWateringFrequency
    )
    run_time_group: RunTimeGroup = _optional_field(default_factory=RunTimeGroup)


class AdvancedProgramDayPatternEnum(_AutoEnum):
    """A value for an advanced watering program day pattern."""

    EVEN = auto()
    ODD = auto()
    MONDAY = auto()
    TUESDAY = auto()
    WEDNESDAY = auto()
    THURSDAY = auto()
    FRIDAY = auto()
    SATURDAY = auto()
    SUNDAY = auto()
    DAYS = auto()


@dataclass
class WateringSettings:
    """Generic settings for a watering program."""

    fixed_watering_adjustment: int = 0
    cycle_and_soak_settings: Optional[CycleAndSoakSettings] = None


@dataclass
class AdvancedWateringSettings(WateringSettings):
    """Advanced watering program settings."""

    advanced_program: Optional[AdvancedProgram] = None


@dataclass
@type_name("Unit")
class TimeRange:
    """Time range units."""

    valid_from: datetime = _optional_field(
        metadata=_timestamp_conversion(), default_factory=datetime
    )
    valid_to: datetime = _optional_field(
        metadata=_timestamp_conversion(), default_factory=datetime
    )


@dataclass
class StandardProgramPeriodicity:
    """Program frequency for a standard program."""

    period: int = 0
    series_start: datetime = field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )


@dataclass
class StandardProgram(Program):
    """A standard watering program."""

    start_times: list[time] = _optional_field(
        metadata=_list_conversion(_time_conversion()), default_factory=list
    )
    time_range: TimeRange = field(default_factory=TimeRange)
    ignore_rain_sensor: bool = False
    days_run: list[DaysOfWeekEnum] = field(default_factory=list)
    standard_program_day_pattern: str = ""
    periodicity: StandardProgramPeriodicity = _optional_field(
        default_factory=StandardProgramPeriodicity
    )


@dataclass
class StandardProgramApplication:
    """A standard watering program application."""

    zone: BaseZone = field(default_factory=BaseZone)
    standard_program: StandardProgram = field(default_factory=StandardProgram)
    run_time_group: RunTimeGroup = field(default_factory=RunTimeGroup)


@dataclass
class StandardWateringSettings(WateringSettings):
    """Standard watering settings."""

    standard_program_applications: list[StandardProgramApplication] = field(
        default_factory=list
    )


@dataclass
class RunStatus:
    """Run status."""

    value: int = _optional_field(default=0)
    label: str = _optional_field(default="")


@dataclass
class ScheduledZoneRun:
    """A scheduled zone run."""

    id: str = ""
    start_time: datetime = field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )
    end_time: datetime = field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )
    normal_duration: timedelta = field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )
    duration: timedelta = field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )
    remaining_time: timedelta = field(
        metadata=_duration_conversion("seconds"), default=timedelta()
    )
    status: RunStatus = field(default_factory=RunStatus)


@dataclass
class ScheduledZoneRuns:
    """Scheduled runs for a zone."""

    summary: str = ""
    current_run: Optional[ScheduledZoneRun] = None
    next_run: Optional[ScheduledZoneRun] = None
    status: Optional[str] = None


@dataclass
class PastZoneRuns:
    """Previous zone runs."""

    last_run: Optional[ScheduledZoneRun] = None
    runs: list[ScheduledZoneRun] = _optional_field(default_factory=list)


@dataclass
class ZoneStatus:
    """A zone's status."""

    relative_water_balance: int = 0
    suspended_until: Optional[datetime] = field(
        metadata=DateTime.conversion(), default=None
    )


@dataclass
class ZoneSuspension:
    """A zone suspension."""

    id: int = 0
    start_time: datetime = _optional_field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )
    end_time: datetime = _optional_field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )


@dataclass
class Zone(BaseZone):
    """A watering zone."""

    watering_settings: Union[
        AdvancedWateringSettings, StandardWateringSettings
    ] = field(default_factory=StandardWateringSettings)
    scheduled_runs: ScheduledZoneRuns = field(default_factory=ScheduledZoneRuns)
    past_runs: PastZoneRuns = field(default_factory=PastZoneRuns)
    status: ZoneStatus = field(default_factory=ZoneStatus)
    suspensions: list[ZoneSuspension] = field(default_factory=list)


@dataclass
class ProgramStartTimeApplication:
    """Application of a start time to a program."""

    all: bool = False
    zones: list[BaseZone] = _optional_field(default_factory=list)


@dataclass
class ProgramStartTime:
    """Start time for a watering program."""

    id: int = 0
    time: time = field(metadata=_time_conversion(), default_factory=time)
    watering_days: list[AdvancedProgramDayPatternEnum] = _optional_field(
        default_factory=list
    )
    application: ProgramStartTimeApplication = field(
        default_factory=ProgramStartTimeApplication
    )


@dataclass
class ControllerFirmware:
    """Information about the controller's firmware."""

    type: str = ""
    version: str = _optional_field(default="")


@dataclass
class ControllerModel:
    """Information about a controller model."""

    name: str = ""
    description: str = ""


@dataclass
class ControllerHardware:
    """Information about a controller's hardware."""

    serial_number: str = _optional_field(default="")
    version: str = _optional_field(default="")
    status: str = _optional_field(default="")
    model: ControllerModel = _optional_field(default_factory=ControllerModel)
    firmware: list[ControllerFirmware] = _optional_field(default_factory=list)


class CustomSensorTypeEnum(_AutoEnum):
    """A value for a sensor type."""

    LEVEL_OPEN = auto()
    LEVEL_CLOSED = auto()
    FLOW = auto()
    THRESHOLD = auto()


@dataclass
class SensorModel:
    """Information about a sensor model."""

    id: int = 0
    name: str = _optional_field(default="")
    active: bool = _optional_field(default=False)
    off_level: int = _optional_field(default=0)
    off_timer: int = _optional_field(default=0)
    delay: timedelta = _optional_field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )
    divisor: float = _optional_field(default=0.0)
    flow_rate: float = _optional_field(default=0.0)
    sensor_type: Optional[CustomSensorTypeEnum] = None


@dataclass
class SensorStatus:
    """Current status of a sensor."""

    water_flow: Optional[LocalizedValueType] = None
    active: bool = _optional_field(default=False)


@dataclass
class SensorFlowSummary:
    """Summary of a sensor's water flow."""

    total_water_volume: LocalizedValueType = _optional_field(
        default_factory=LocalizedValueType
    )


@dataclass
class Sensor:
    """A sensor connected to a controller."""

    id: int = 0
    name: str = ""
    model: SensorModel = field(default_factory=SensorModel)
    status: SensorStatus = field(default_factory=SensorStatus)


@dataclass
class _WaterTime:
    """A water time duration."""

    value: timedelta = _optional_field(
        metadata=_duration_conversion("minutes"), default=timedelta()
    )


@dataclass
class ActualWaterTime(_WaterTime):
    """An actual water time duration."""


@dataclass
class NormalWaterTime(_WaterTime):
    """A normal water time duration."""


@dataclass
class ControllerStatus:
    """Current status of a controller."""

    summary: str = ""
    online: bool = False
    actual_water_time: ActualWaterTime = _optional_field(
        default_factory=ActualWaterTime
    )
    normal_water_time: NormalWaterTime = _optional_field(
        default_factory=NormalWaterTime
    )
    last_contact: Optional[DateTime] = None


@dataclass
class RunStatusType:
    value: int = 0
    label: str = ""


@dataclass
class RunStopReasonType:
    finished_normally: bool = False
    description: list[str] = field(default_factory=list)


@dataclass
@type_name("RunEventType")
class RunEvent:
    """A Hydrawise run event type."""

    id: str = ""
    zone: BaseZone = field(default_factory=BaseZone)
    standard_program: StandardProgramRef = _optional_field(
        default_factory=StandardProgramRef
    )
    advanced_program: AdvancedProgramRef = _optional_field(
        default_factory=AdvancedProgramRef
    )
    reported_start_time: Optional[datetime] = field(
        metadata=DateTime.conversion(), default=None
    )
    reported_end_time: Optional[datetime] = field(
        metadata=DateTime.conversion(), default=None
    )
    reported_duration: timedelta = _optional_field(
        metadata=_duration_conversion("seconds"), default=timedelta()
    )
    reported_status: RunStatusType = _optional_field(default_factory=RunStatusType)
    reported_water_usage: LocalizedValueType = _optional_field(
        default_factory=LocalizedValueType
    )
    reported_stop_reason: RunStopReasonType = _optional_field(
        default_factory=RunStopReasonType
    )
    reported_current: LocalizedValueType = _optional_field(
        default_factory=LocalizedValueType
    )


@dataclass
class WateringReportEntry:
    """A Hydrawise watering report entry."""

    run_event: RunEvent = _optional_field(
        default_factory=RunEvent, metadata=fall_back_on_default
    )


@dataclass
class Controller:
    """A Hydrawise controller."""

    id: int = 0
    name: str = _optional_field(default="")
    software_version: str = _optional_field(default="")
    hardware: ControllerHardware = field(default_factory=ControllerHardware)
    last_contact_time: datetime = _optional_field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )
    last_action: datetime = _optional_field(
        metadata=DateTime.conversion(), default_factory=default_datetime
    )
    online: bool = _optional_field(default=False)
    sensors: list[Sensor] = _optional_field(default_factory=list)
    zones: list[Zone] = _optional_field(default_factory=list)
    permitted_program_start_times: list[ProgramStartTime] = _optional_field(
        default_factory=list
    )
    status: Optional[ControllerStatus] = None


@dataclass
class User:
    """A Hydrawise user account."""

    id: int = 0
    customer_id: int = 0
    name: str = ""
    email: str = _optional_field(default="")
    controllers: list[Controller] = _optional_field(default_factory=list)


class DaysOfWeekEnum(_AutoEnum):
    """All days of the week."""

    SUNDAY = auto()
    MONDAY = auto()
    TUESDAY = auto()
    WEDNESDAY = auto()
    THURSDAY = auto()
    FRIDAY = auto()
    SATURDAY = auto()


@dataclass
class ControllerWaterUseSummary:
    """Water use summary for a controller.

    Active use means water use during a scheduled or manual zone run.
    Inactive use means water use when no zone was actively running. This can happen when
    faucets (i.e., garden hoses) are installed downstream of the flow meter.
    """

    _pydrawise_type = True

    total_use: float = 0.0
    total_active_use: float = 0.0
    total_inactive_use: float = 0.0
    active_use_by_zone_id: dict[int, float] = field(default_factory=dict)
    unit: str = ""



from datetime import datetime, timedelta
from unittest.mock import create_autospec, patch

from gql import Client
from gql.client import AsyncClientSession
from graphql import print_ast
import pytest
from pytest import fixture

from pydrawise.auth import Auth
from pydrawise.client import Hydrawise
from pydrawise.schema import Controller, Sensor, Zone, ZoneSuspension
from pydrawise.schema_utils import deserialize


@fixture
def mock_auth():
    mock_auth = create_autospec(Auth, spec_set=True, instance=True)
    mock_auth.token.return_value = "__token__"
    yield mock_auth


@fixture
def mock_session():
    yield create_autospec(AsyncClientSession, spec_set=True, instance=True)


@fixture
def mock_client(mock_session):
    client = create_autospec(Client, spec_set=True, instance=True)
    client.__aenter__.return_value = mock_session
    yield client


@fixture
def api(mock_auth, mock_client):
    api = Hydrawise(mock_auth)
    with patch.object(api, "_client", return_value=mock_client):
        yield api


@fixture
def rain_sensor_json():
    yield {
        "id": 337844,
        "name": "Rain sensor ",
        "model": {
            "id": 3318,
            "name": "Rain Sensor (normally closed wire)",
            "active": True,
            "offLevel": 1,
            "offTimer": 0,
            "delay": 0,
            "divisor": 0,
            "flowRate": 0,
            "sensorType": "LEVEL_CLOSED",
        },
        "status": {
            "waterFlow": None,
            "active": False,
        },
    }


@fixture
def flow_sensor_json():
    yield {
        "id": 337845,
        "name": "Flow meter",
        "model": {
            "id": 3324,
            "name": "1, 1½ or 2 inch NPT Flow Meter",
            "active": True,
            "offLevel": 0,
            "offTimer": 0,
            "delay": 0,
            "divisor": 0.52834,
            "flowRate": 3.7854,
            "sensorType": "FLOW",
        },
        "status": {
            "waterFlow": {
                "value": 542.0042035155608,
                "unit": "gal",
            },
            "active": None,
        },
    }


@fixture
def flow_summary_json(request):
    if request.param:
        yield {"totalWaterVolume": {"value": 23134.67952992029, "unit": "gal"}}
    else:
        yield None


@fixture
def controller_json(rain_sensor_json, flow_sensor_json):
    yield {
        "id": 9876,
        "name": "Main Controller",
        "softwareVersion": "s0",
        "hardware": {
            "serialNumber": "A0B1C2D3",
            "version": "1.0",
            "status": "All good!",
            "model": {
                "name": "HPC 10",
                "description": "HPC 10 Station Controller",
            },
            "firmware": [{"type": "A", "version": "1.0"}],
        },
        "lastContactTime": {
            "timestamp": 1672531200,
            "value": "Sun, 01 Jan 23 00:12:00",
        },
        "lastAction": {
            "timestamp": 1672531200,
            "value": "Sun, 01 Jan 23 00:12:00",
        },
        "online": True,
        "sensors": [rain_sensor_json, flow_sensor_json],
        "permittedProgramStartTimes": [],
        "status": {
            "summary": "All good!",
            "online": True,
            "actualWaterTime": {"value": 10},
            "normalWaterTime": {"value": 10},
            "lastContact": {
                "timestamp": 1672531200,
                "value": "Sun, 01 Jan 23 00:12:00",
            },
        },
    }


@fixture
def zone_json():
    yield {
        "id": 1,
        "number": {
            "value": 1,
            "label": "One",
        },
        "name": "Zone A",
        "wateringSettings": {
            "fixedWateringAdjustment": 100,
            "cycleAndSoakSettings": None,
            "advancedProgram": {
                "id": 4729361,
                "name": "",
                "schedulingMethod": {"value": 0, "label": "Time Based"},
                "monthlyWateringAdjustments": [
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                    100,
                ],
                "appliesToZones": [
                    {
                        "id": 5955343,
                        "number": {"value": 1, "label": "Zone 1"},
                        "name": "Front Lawn",
                    }
                ],
                "zoneSpecific": True,
                "advancedProgramId": 5655942,
                "wateringFrequency": {
                    "label": "Frequency",
                    "period": {
                        "value": None,
                        "label": "Every Program Start Time",
                    },
                    "description": "Every Program Start Time unless modified by your Watering Triggers",
                },
                "runTimeGroup": {
                    "id": 49923604,
                    "name": None,
                    "duration": 20,
                },
            },
        },
        "scheduledRuns": {
            "summary": "",
            "currentRun": None,
            "nextRun": None,
            "status": None,
        },
        "pastRuns": {"lastRun": None, "runs": []},
        "status": {
            "relativeWaterBalance": 0,
            "suspendedUntil": {
                "timestamp": 1672531200,
                "value": "Sun, 01 Jan 23 00:12:00",
            },
        },
        "suspensions": [],
    }


@fixture
def watering_report_json():
    yield {
        "watering": [
            {
                "runEvent": {
                    "id": "35220026902",
                    "zone": {
                        "id": 5955343,
                        "number": {"value": 1, "label": "Zone 1"},
                        "name": "Front Lawn",
                    },
                    "standardProgram": {
                        "id": 343434,
                        "name": "",
                    },
                    "advancedProgram": {"id": 4729361, "name": ""},
                    "reportedStartTime": {
                        "value": "Fri, 01 Dec 23 04:00:00 -0800",
                        "timestamp": 1701432000,
                    },
                    "reportedEndTime": {
                        "value": "Fri, 01 Dec 23 04:20:00 -0800",
                        "timestamp": 1701433200,
                    },
                    "reportedDuration": 1200,
                    "reportedStatus": {
                        "value": 1,
                        "label": "Normal watering cycle",
                    },
                    "reportedWaterUsage": {
                        "value": 34.000263855044786,
                        "unit": "gal",
                    },
                    "reportedStopReason": {
                        "finishedNormally": True,
                        "description": ["Finished normally"],
                    },
                    "reportedCurrent": {"value": 280, "unit": "mA"},
                }
            },
            {
                "runEvent": {
                    "id": "35220026903",
                    "zone": {
                        "id": 5955345,
                        "number": {"value": 2, "label": "Zone 2"},
                        "name": "Front Trees",
                    },
                    "standardProgram": None,
                    "advancedProgram": {"id": 4729362, "name": ""},
                    "reportedStartTime": {
                        "value": "Fri, 01 Nov 23 04:19:59 -0800",
                        "timestamp": 1698797999,
                    },
                    "reportedEndTime": {
                        "value": "Fri, 01 Nov 23 04:39:59 -0800",
                        "timestamp": 1698799199,
                    },
                    "reportedDuration": 1200,
                    "reportedStatus": {
                        "value": 1,
                        "label": "Normal watering cycle",
                    },
                    "reportedWaterUsage": {
                        "value": 49.00048126864295,
                        "unit": "gal",
                    },
                    "reportedStopReason": {
                        "finishedNormally": True,
                        "description": ["Finished normally"],
                    },
                    "reportedCurrent": {"value": 280, "unit": "mA"},
                }
            },
        ]
    }


async def test_get_user(api: Hydrawise, mock_session, controller_json, zone_json):
    controller_json["zones"] = [zone_json]
    mock_session.execute.return_value = {
        "me": {
            "id": 1234,
            "customerId": 1,
            "name": "My Name",
            "email": "me@asdf.com",
            "controllers": [controller_json],
        }
    }
    user = await api.get_user()
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "controllers {" in query
    assert query.count("zones {") == 2
    assert user.id == 1234
    assert user.name == "My Name"
    assert user.email == "me@asdf.com"
    assert len(user.controllers) == 1
    assert len(user.controllers[0].zones) == 1


async def test_get_user_no_zones(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {
        "me": {
            "id": 1234,
            "customerId": 1,
            "name": "My Name",
            "email": "me@asdf.com",
            "controllers": [controller_json],
        }
    }
    user = await api.get_user(fetch_zones=False)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "controllers {" in query
    assert query.count("zones {") == 1
    assert user.id == 1234
    assert user.name == "My Name"
    assert user.email == "me@asdf.com"
    assert len(user.controllers) == 1
    assert len(user.controllers[0].zones) == 0


async def test_get_controllers(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"me": {"controllers": [controller_json]}}
    [controller] = await api.get_controllers()
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert query.count("zones {") == 2
    assert controller.last_contact_time == datetime(2023, 1, 1, 0, 0, 0)
    assert controller.last_action == datetime(2023, 1, 1, 0, 0, 0)
    assert controller.status.actual_water_time.value == timedelta(minutes=10)


async def test_get_controller(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"controller": controller_json}
    controller = await api.get_controller(9876)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "controller(" in query
    assert "controllerId: 9876" in query
    assert query.count("zones {") == 2

    assert controller.last_contact_time == datetime(2023, 1, 1, 0, 0, 0)
    assert controller.last_action == datetime(2023, 1, 1, 0, 0, 0)
    assert controller.status.actual_water_time.value == timedelta(minutes=10)


async def test_get_zones(api: Hydrawise, mock_session, controller_json, zone_json):
    mock_session.execute.return_value = {"controller": {"zones": [zone_json]}}
    ctrl = deserialize(Controller, controller_json)
    [zone] = await api.get_zones(ctrl)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "controller(" in query
    assert "controllerId: 9876" in query


async def test_get_zone(api: Hydrawise, mock_session, zone_json):
    mock_session.execute.return_value = {"zone": zone_json}
    zone = await api.get_zone(1)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "zone(" in query
    assert "zoneId: 1" in query


async def test_start_zone(api: Hydrawise, mock_session, zone_json):
    mock_session.execute.return_value = {"startZone": {"status": "OK"}}
    zone = deserialize(Zone, zone_json)
    await api.start_zone(zone, custom_run_duration=10)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "startZone(" in query
    assert "zoneId: 1" in query
    assert "markRunAsScheduled: false" in query
    assert "customRunDuration: 10" in query


async def test_stop_zone(api: Hydrawise, mock_session, zone_json):
    mock_session.execute.return_value = {"stopZone": {"status": "OK"}}
    zone = deserialize(Zone, zone_json)
    await api.stop_zone(zone)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "stopZone(" in query
    assert "zoneId: 1" in query


async def test_start_all_zones(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"startAllZones": {"status": "OK"}}
    ctrl = deserialize(Controller, controller_json)
    await api.start_all_zones(ctrl, custom_run_duration=10)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "startAllZones(" in query
    assert "controllerId: 9876" in query
    assert "markRunAsScheduled: false" in query
    assert "customRunDuration: 10" in query


async def test_stop_all_zones(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"stopAllZones": {"status": "OK"}}
    ctrl = deserialize(Controller, controller_json)
    await api.stop_all_zones(ctrl)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "stopAllZones(" in query
    assert "controllerId: 9876" in query


async def test_suspend_zone(api: Hydrawise, mock_session, zone_json):
    mock_session.execute.return_value = {"suspendZone": {"status": "OK"}}
    zone = deserialize(Zone, zone_json)
    await api.suspend_zone(zone, until=datetime(2023, 1, 1, 0, 0, 0))
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "suspendZone(" in query
    assert "zoneId: 1" in query
    assert 'until: "Sun, 01 Jan 23 00:12:00 +0000"' in query


async def test_resume_zone(api: Hydrawise, mock_session, zone_json):
    mock_session.execute.return_value = {"resumeZone": {"status": "OK"}}
    zone = deserialize(Zone, zone_json)
    await api.resume_zone(zone)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "resumeZone(" in query
    assert "zoneId: 1" in query


async def test_suspend_all_zones(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"suspendAllZones": {"status": "OK"}}
    ctrl = deserialize(Controller, controller_json)
    await api.suspend_all_zones(ctrl, until=datetime(2023, 1, 1, 0, 0, 0))
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "suspendAllZones(" in query
    assert "controllerId: 9876" in query
    assert 'until: "Sun, 01 Jan 23 00:12:00 +0000"' in query


async def test_resume_all_zones(api: Hydrawise, mock_session, controller_json):
    mock_session.execute.return_value = {"resumeAllZones": {"status": "OK"}}
    ctrl = deserialize(Controller, controller_json)
    await api.resume_all_zones(ctrl)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "resumeAllZones(" in query
    assert "controllerId: 9876" in query


async def test_delete_zone_suspension(api: Hydrawise, mock_session):
    mock_session.execute.return_value = {"deleteZoneSuspension": True}
    suspension = ZoneSuspension(
        id=2222,
        start_time=datetime(2023, 1, 1, 0, 0, 0),
        end_time=datetime(2023, 1, 2, 0, 0, 0),
    )
    await api.delete_zone_suspension(suspension)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "deleteZoneSuspension(" in query
    assert "id: 2222" in query


async def test_get_sensors(
    api: Hydrawise,
    mock_session,
    rain_sensor_json,
    flow_sensor_json,
    controller_json,
):
    mock_session.execute.return_value = {
        "controller": {"sensors": [rain_sensor_json, flow_sensor_json]}
    }
    ctrl = deserialize(Controller, controller_json)
    sensors = await api.get_sensors(ctrl)
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "sensors {" in query


@pytest.mark.parametrize("flow_summary_json", (True, False), indirect=True)
async def test_get_water_flow_summary(
    api: Hydrawise,
    mock_session,
    controller_json,
    flow_sensor_json,
    flow_summary_json,
):
    mock_session.execute.return_value = {
        "controller": {
            "sensors": [flow_sensor_json | {"flowSummary": flow_summary_json}]
        }
    }

    ctrl = deserialize(Controller, controller_json)
    sensor = deserialize(Sensor, flow_sensor_json)
    water_flow_summary = await api.get_water_flow_summary(
        ctrl,
        sensor,
        datetime(2023, 11, 1, 0, 0, 0),
        datetime(2023, 11, 30, 0, 0, 0),
    )
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "flowSummary(" in query


async def test_get_watering_report(
    api: Hydrawise, mock_session, controller_json, watering_report_json
):
    mock_session.execute.return_value = {
        "controller": {"reports": watering_report_json}
    }
    ctrl = deserialize(Controller, controller_json)
    report = await api.get_watering_report(
        ctrl, datetime(2023, 12, 1, 0, 0, 0), datetime(2023, 12, 4, 0, 0, 0)
    )
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "reports" in query
    assert "watering" in query
    assert len(report) == 1


@pytest.mark.parametrize("flow_summary_json", (True, False), indirect=True)
async def test_get_water_use_summary(
    api: Hydrawise,
    mock_session,
    controller_json,
    watering_report_json,
    flow_sensor_json,
    flow_summary_json,
):
    mock_session.execute.return_value = {
        "controller": {
            "reports": watering_report_json,
            "sensors": [flow_sensor_json | {"flowSummary": flow_summary_json}],
        }
    }
    ctrl = deserialize(Controller, controller_json)
    summary = await api.get_water_use_summary(
        ctrl, datetime(2023, 12, 1, 0, 0, 0), datetime(2023, 12, 4, 0, 0, 0)
    )
    mock_session.execute.assert_awaited_once()
    [selector] = mock_session.execute.await_args.args
    query = print_ast(selector)
    assert "reports" in query
    assert "watering" in query
    assert "flowSummary(" in query
    assert summary.active_use_by_zone_id[5955343] == 34.000263855044786
    assert summary.total_active_use == 34.000263855044786
    assert summary.total_inactive_use == (
        23100.679266065246 if flow_summary_json else 0.0
    )
    assert summary.unit == "gal"



"""API for interacting with Hydrawise sprinkler controllers."""

from .auth import Auth
from .base import HydrawiseBase
from .client import Hydrawise
from .exceptions import (
    Error,
    MutationError,
    NotAuthenticatedError,
    NotAuthorizedError,
    UnknownError,
)
from .schema import Controller, Sensor, User, Zone

__all__ = (
    "Auth",
    "Controller",
    "Error",
    "Hydrawise",
    "MutationError",
    "NotAuthenticatedError",
    "NotAuthorizedError",
    "Sensor",
    "UnknownError",
    "User",
    "Zone",
)



"""Client library for interacting with Hydrawise's cloud API."""

from datetime import datetime
from functools import cache
from importlib import resources
import logging

from gql import Client
from gql.dsl import DSLField, DSLMutation, DSLQuery, DSLSchema, DSLSelectable, dsl_gql
from gql.transport.aiohttp import AIOHTTPTransport, log as gql_log
from graphql import GraphQLSchema, build_ast_schema, parse

from .auth import Auth
from .base import HydrawiseBase
from .exceptions import MutationError
from .schema import (
    Controller,
    ControllerWaterUseSummary,
    DateTime,
    LocalizedValueType,
    Sensor,
    SensorFlowSummary,
    StatusCodeAndSummary,
    User,
    WateringReportEntry,
    Zone,
    ZoneSuspension,
)
from .schema_utils import deserialize, get_selectors

# GQL is quite chatty in logs by default.
gql_log.setLevel(logging.ERROR)

API_URL = "https://app.hydrawise.com/api/v2/graph"


@cache
def _get_schema() -> GraphQLSchema:
    schema_text = resources.files(__package__).joinpath("hydrawise.graphql").read_text()
    return build_ast_schema(parse(schema_text))


def _prune_watering_report_entries(
    entries: list[WateringReportEntry], start: datetime, end: datetime
) -> list[WateringReportEntry]:
    """Prune watering report entries to make sure they all fall inside the [start, end] time interval.

    The call to watering() can return events outside of the provided time interval.
    Filter out events that happen before or after the provided time interval.
    """
    return list(
        filter(
            lambda entry: entry.run_event is not None
            and entry.run_event.reported_start_time is not None
            and entry.run_event.reported_end_time is not None
            and (
                (
                    start.timestamp()
                    <= entry.run_event.reported_start_time.timestamp()
                    <= end.timestamp()
                )
                or (
                    start.timestamp()
                    <= entry.run_event.reported_end_time.timestamp()
                    <= end.timestamp()
                )
            ),
            entries,
        )
    )


class Hydrawise(HydrawiseBase):
    """Client library for interacting with Hydrawise sprinkler controllers.

    Should be instantiated with an Auth object that handles authentication and low-level transport.
    """

    def __init__(self, auth: Auth) -> None:
        """Initializes the client.

        :param auth: Handles authentication and transport.
        """
        self._auth = auth
        self._schema = DSLSchema(_get_schema())

    async def _client(self) -> Client:
        headers = {"Authorization": await self._auth.token()}
        transport = AIOHTTPTransport(url=API_URL, headers=headers)
        return Client(transport=transport, parse_results=True)

    async def _query(self, selector: DSLSelectable) -> dict:
        async with await self._client() as session:
            return await session.execute(dsl_gql(DSLQuery(selector)))

    async def _mutation(self, selector: DSLField) -> None:
        async with await self._client() as session:
            result = await session.execute(dsl_gql(DSLMutation(selector)))
            resp = result[selector.name]
            if isinstance(resp, dict):
                if resp["status"] != "OK":
                    raise MutationError(resp["summary"])
                return
            elif not resp:
                # Assume bool response
                raise MutationError

    async def get_user(self, fetch_zones: bool = True) -> User:
        """Retrieves the currently authenticated user.

        :param fetch_zones: Not used in this implementation.
        :rtype: User
        """
        skip = [] if fetch_zones else ["controllers.zones"]
        selector = self._schema.Query.me.select(
            *get_selectors(self._schema, User, skip)
        )
        result = await self._query(selector)
        return deserialize(User, result["me"])

    async def get_controllers(self) -> list[Controller]:
        """Retrieves all controllers associated with the currently authenticated user.

        :rtype: list[Controller]
        """
        selector = self._schema.Query.me.select(
            self._schema.User.controllers.select(
                *get_selectors(self._schema, Controller)
            ),
        )
        result = await self._query(selector)
        return deserialize(list[Controller], result["me"]["controllers"])

    async def get_controller(self, controller_id: int) -> Controller:
        """Retrieves a single controller by its unique identifier.

        :param controller_id: Unique identifier for the controller to retrieve.
        :rtype: Controller
        """
        selector = self._schema.Query.controller(controllerId=controller_id).select(
            *get_selectors(self._schema, Controller),
        )
        result = await self._query(selector)
        return deserialize(Controller, result["controller"])

    async def get_zones(self, controller: Controller) -> list[Zone]:
        """Retrieves zones associated with the given controller.

        :param controller: Controller whose zones to fetch.
        :rtype: list[Zone]
        """
        selector = self._schema.Query.controller(controllerId=controller.id).select(
            self._schema.Controller.zones.select(*get_selectors(self._schema, Zone)),
        )
        result = await self._query(selector)
        return deserialize(list[Zone], result["controller"]["zones"])

    async def get_zone(self, zone_id: int) -> Zone:
        """Retrieves a zone by its unique identifier.

        :param zone_id: The zone's unique identifier.
        :rtype: Zone
        """
        selector = self._schema.Query.zone(zoneId=zone_id).select(
            *get_selectors(self._schema, Zone)
        )
        result = await self._query(selector)
        return deserialize(Zone, result["zone"])

    async def start_zone(
        self,
        zone: Zone,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts a zone's run cycle.

        :param zone: The zone to start.
        :param mark_run_as_scheduled: Whether to mark the zone as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zone. If not
            specified (or zero), will run for its default configured time.
        """
        kwargs = {
            "zoneId": zone.id,
            "markRunAsScheduled": mark_run_as_scheduled,
        }
        if custom_run_duration > 0:
            kwargs["customRunDuration"] = custom_run_duration

        selector = self._schema.Mutation.startZone.args(**kwargs).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def stop_zone(self, zone: Zone) -> None:
        """Stops a zone.

        :param zone: The zone to stop.
        """
        selector = self._schema.Mutation.stopZone.args(zoneId=zone.id).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def start_all_zones(
        self,
        controller: Controller,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts all zones attached to a controller.

        :param controller: The controller whose zones to start.
        :param mark_run_as_scheduled: Whether to mark the zones as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zones. If not
            specified (or zero), will run for each zone's default configured time.
        """
        kwargs = {
            "controllerId": controller.id,
            "markRunAsScheduled": mark_run_as_scheduled,
        }
        if custom_run_duration > 0:
            kwargs["customRunDuration"] = custom_run_duration

        selector = self._schema.Mutation.startAllZones.args(**kwargs).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def stop_all_zones(self, controller: Controller) -> None:
        """Stops all zones attached to a controller.

        :param controller: The controller whose zones to stop.
        """
        selector = self._schema.Mutation.stopAllZones.args(
            controllerId=controller.id
        ).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def suspend_zone(self, zone: Zone, until: datetime) -> None:
        """Suspends a zone's schedule.

        :param zone: The zone to suspend.
        :param until: When the suspension should end.
        """
        selector = self._schema.Mutation.suspendZone.args(
            zoneId=zone.id,
            until=DateTime.to_json(until).value,
        ).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def resume_zone(self, zone: Zone) -> None:
        """Resumes a zone's schedule.

        :param zone: The zone whose schedule to resume.
        """
        selector = self._schema.Mutation.resumeZone.args(zoneId=zone.id).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def suspend_all_zones(self, controller: Controller, until: datetime) -> None:
        """Suspends the schedule of all zones attached to a given controller.

        :param controller: The controller whose zones to suspend.
        :param until: When the suspension should end.
        """
        selector = self._schema.Mutation.suspendAllZones.args(
            controllerId=controller.id,
            until=DateTime.to_json(until).value,
        ).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def resume_all_zones(self, controller: Controller) -> None:
        """Resumes the schedule of all zones attached to the given controller.

        :param controller: The controller whose zones to resume.
        """
        selector = self._schema.Mutation.resumeAllZones.args(
            controllerId=controller.id
        ).select(
            *get_selectors(self._schema, StatusCodeAndSummary),
        )
        await self._mutation(selector)

    async def delete_zone_suspension(self, suspension: ZoneSuspension) -> None:
        """Removes a specific zone suspension.

        Useful when there are multiple suspensions for a zone in effect.

        :param suspension: The suspension to delete.
        """
        selector = self._schema.Mutation.deleteZoneSuspension.args(
            id=suspension.id
        ).select()
        await self._mutation(selector)

    async def get_sensors(self, controller: Controller) -> list[Sensor]:
        """Retrieves sensors associated with the given controller.

        :param controller: Controller whose sensors to fetch.
        :rtype: list[Sensor]
        """
        selector = self._schema.Query.controller(controllerId=controller.id).select(
            self._schema.Controller.sensors.select(
                *get_selectors(self._schema, Sensor)
            ),
        )

        result = await self._query(selector)
        return deserialize(list[Sensor], result["controller"]["sensors"])

    async def get_water_flow_summary(
        self, controller: Controller, sensor: Sensor, start: datetime, end: datetime
    ) -> SensorFlowSummary:
        """Retrieves the water flow summary for a given sensor.

        :param controller: Controller that controls the sensor.
        :param sensor: Sensor for which a water flow summary is fetched.
        :param start:
        :param end:
        :rtype: list[Sensor]
        """
        selector = self._schema.Query.controller(controllerId=controller.id).select(
            self._schema.Controller.sensors.select(
                *get_selectors(self._schema, Sensor),
                self._schema.Sensor.flowSummary(
                    start=int(start.timestamp()),
                    end=int(end.timestamp()),
                ).select(*get_selectors(self._schema, SensorFlowSummary)),
            ),
        )

        result = await self._query(selector)

        # There is no way to query for one particular sensor through GraphQL. We need
        # to filter here instead. This should not really be a performance problem in
        # practice since it is unlikely for a controller to have more than one water
        # sensor.
        sensors = list(
            filter(lambda s: s["id"] == sensor.id, result["controller"]["sensors"])
        )
        if len(sensors) == 0:
            raise ValueError(f"Sensor with id={sensor.id} not found")
        if "flowSummary" not in sensors[0]:
            raise ValueError(
                f"Sensor with id={sensor.id} does not have any flow information"
            )
        if sensors[0]["flowSummary"] is None:
            return SensorFlowSummary(total_water_volume=LocalizedValueType(0.0, "gal"))
        return deserialize(SensorFlowSummary, sensors[0]["flowSummary"])

    async def get_watering_report(
        self, controller: Controller, start: datetime, end: datetime
    ) -> list[WateringReportEntry]:
        """Retrieves a watering report for the given controller and time period.

        :param controller: The controller whose watering report to generate.
        :param start: Start time.
        :param end: End time."""
        selector = self._schema.Query.controller(controllerId=controller.id).select(
            self._schema.Controller.reports.select(
                self._schema.Reports.watering(
                    **{
                        "from": int(start.timestamp()),
                        "until": int(end.timestamp()),
                    }
                ).select(*get_selectors(self._schema, WateringReportEntry)),
            ),
        )
        result = await self._query(selector)
        return _prune_watering_report_entries(
            deserialize(
                list[WateringReportEntry], result["controller"]["reports"]["watering"]
            ),
            start,
            end,
        )

    async def get_water_use_summary(
        self, controller: Controller, start: datetime, end: datetime
    ) -> ControllerWaterUseSummary:
        """Calculate the water use for the given controller and time period.

        :param controller: The controller whose water use to report.
        :param start: Start time
        :param end: End time."""
        selector = self._schema.Query.controller(controllerId=controller.id).select(
            self._schema.Controller.sensors.select(
                *get_selectors(self._schema, Sensor),
                self._schema.Sensor.flowSummary(
                    start=int(start.timestamp()),
                    end=int(end.timestamp()),
                ).select(*get_selectors(self._schema, SensorFlowSummary)),
            ),
            self._schema.Controller.reports.select(
                self._schema.Reports.watering(
                    **{
                        "from": int(start.timestamp()),
                        "until": int(end.timestamp()),
                    }
                ).select(*get_selectors(self._schema, WateringReportEntry)),
            ),
        )
        result = await self._query(selector)
        summary = ControllerWaterUseSummary()

        # watering report entries
        entries = _prune_watering_report_entries(
            deserialize(
                list[WateringReportEntry], result["controller"]["reports"]["watering"]
            ),
            start,
            end,
        )

        # total active water use
        for entry in entries:
            if (
                entry.run_event is not None
                and entry.run_event.zone is not None
                and entry.run_event.reported_water_usage is not None
            ):
                active_use = entry.run_event.reported_water_usage.value
                if summary.unit == "":
                    summary.unit = entry.run_event.reported_water_usage.unit
                summary.total_active_use += active_use
                summary.active_use_by_zone_id.setdefault(entry.run_event.zone.id, 0)
                summary.active_use_by_zone_id[entry.run_event.zone.id] += active_use

        # total active and inactive water use
        for sensor in result["controller"]["sensors"]:
            if (
                "FLOW" in sensor["model"]["sensorType"]
                and "flowSummary" in sensor
                and (flow_summary := sensor["flowSummary"]) is not None
            ):
                summary.total_use += flow_summary["totalWaterVolume"]["value"]
                if summary.unit == "":
                    summary.unit = flow_summary["totalWaterVolume"]["unit"]

        # Correct for inaccuracies. The watering report and flow summaries are not always
        # updated with the same frequency.
        if summary.total_use > summary.total_active_use:
            summary.total_inactive_use = summary.total_use - summary.total_active_use
        else:
            summary.total_use = summary.total_active_use

        return summary



from dataclasses import fields, is_dataclass
from pathlib import Path
from string import ascii_lowercase

from apischema.metadata.keys import CONVERSION_METADATA, FALL_BACK_ON_DEFAULT_METADATA
from apischema.type_names import get_type_name
from apischema.utils import to_camel_case
from graphql import build_schema
from graphql.type import GraphQLBoolean, GraphQLInt, GraphQLNonNull, GraphQLString

from pydrawise import schema as _schema


def test_valid_schema():
    root = Path(__file__).parent.parent
    with open(Path(f"{root}/pydrawise/hydrawise.graphql")) as f:
        schema_str = f.read()

    gql_schema = build_schema(schema_str)

    for k, v in vars(_schema).items():
        if k.startswith("_") or k.startswith(ascii_lowercase):
            # Ignore private types
            continue
        if getattr(v, "_pydrawise_type", False):
            # Ignore internal types
            continue
        if not is_dataclass(v):
            # Only look at dataclass types
            continue
        if not (name := get_type_name(v).graphql):
            # Ignore types that are not graphql types.
            continue

        st = gql_schema.get_type(name)
        assert st is not None, f"{name} not found in schema"
        for f in fields(v):
            if f.name.startswith("_"):
                # Ignore private fields.
                continue
            fname = to_camel_case(f.name)
            assert fname in st.fields, f"{name}.{f.name} is not a valid field"
            sf = st.fields[fname]

            want_optional = False
            if isinstance(sf.type, GraphQLNonNull):
                stype = sf.type.of_type
            else:
                stype = sf.type
                want_optional = FALL_BACK_ON_DEFAULT_METADATA not in f.metadata
                assert (
                    FALL_BACK_ON_DEFAULT_METADATA in f.metadata or "Optional" in f.type
                ), f"{name}.{f.name} should be optional"

            type_map = {
                GraphQLBoolean: "bool",
                GraphQLInt: "int",
                GraphQLString: "str",
            }
            if stype not in type_map:
                continue
            want_type = type_map[stype]
            if want_optional:
                want_type = f"Optional[{want_type}]"
            got_type = f.type
            if md := f.metadata.get(CONVERSION_METADATA):
                assert md.deserialization is not None
                # TODO: Validate the source type.
                # got_type = md.deserialization.source
                continue
            assert (
                got_type == want_type
            ), f"{name}.{f.name} is {got_type}, want {want_type}"



""""API for interacting with Hydrawise sprinkler controllers.

This library should remain compatible with https://github.com/ptcryan/hydrawiser.
 """

from datetime import datetime, timedelta
import time
from typing import Any

import aiohttp
import requests

from .base import HydrawiseBase
from .exceptions import NotInitializedError, UnknownError
from .schema import (
    Controller,
    ControllerHardware,
    ScheduledZoneRun,
    ScheduledZoneRuns,
    User,
    Zone,
    ZoneStatus,
    ZoneSuspension,
)

_BASE_URL = "https://api.hydrawise.com/api/v1"
_TIMEOUT = 10  # seconds


class LegacyHydrawiseAsync(HydrawiseBase):
    """Async client library for interacting with the Hydrawise v1 API.

    This should remain compatible with client.Hydrawise.
    """

    def __init__(self, user_token: str) -> None:
        self._api_key = user_token

    async def _get(self, path: str, **kwargs) -> dict:
        url = f"{_BASE_URL}/{path}"
        params = {"api_key": self._api_key}
        params.update(kwargs)
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params, timeout=_TIMEOUT) as resp:
                resp.raise_for_status()
                return await resp.json()

    async def get_user(self, fetch_zones: bool = True) -> User:
        """Retrieves the currently authenticated user.

        :param fetch_zones: When True, also fetch zones.
        :rtype: User
        """
        resp_json = await self._get("customerdetails.php")
        user = User(
            id=0,
            customer_id=resp_json["customer_id"],
            name="",
            email="",
            controllers=[_controller_from_json(c) for c in resp_json["controllers"]],
        )
        if fetch_zones:
            for controller in user.controllers:
                controller.zones = await self.get_zones(controller)
        return user

    async def get_controllers(self) -> list[Controller]:
        """Retrieves all controllers associated with the currently authenticated user.

        :rtype: list[Controller]
        """
        resp_json = await self._get("customerdetails.php", type="controllers")
        controllers = [_controller_from_json(c) for c in resp_json["controllers"]]
        for controller in controllers:
            controller.zones = await self.get_zones(controller)
        return controllers

    async def get_controller(self, controller_id: int) -> Controller:
        """Retrieves a single controller by its unique identifier.

        :param controller_id: Unique identifier for the controller to retrieve.
        :rtype: Controller
        """
        _ = controller_id  # unused
        raise NotImplementedError

    async def get_zones(self, controller: Controller) -> list[Zone]:
        """Retrieves zones associated with the given controller.

        :param controller: Controller whose zones to fetch.
        :rtype: list[Zone]
        """
        resp_json = await self._get("statusschedule.php", controller_id=controller.id)
        return [_zone_from_json(z) for z in resp_json["relays"]]

    async def get_zone(self, zone_id: int) -> Zone:
        """Retrieves a zone by its unique identifier.

        :param zone_id: The zone's unique identifier.
        :rtype: Zone
        """
        _ = zone_id  # unused
        raise NotImplementedError

    async def start_zone(
        self,
        zone: Zone,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts a zone's run cycle.

        :param zone: The zone to start.
        :param mark_run_as_scheduled: Whether to mark the zone as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zone. If not
            specified (or zero), will run for its default configured time.
        """
        _ = mark_run_as_scheduled  # unused.
        params = {
            "action": "run",
            "relay_id": zone.id,
            "period_id": 999,
        }
        if custom_run_duration > 0:
            params["custom"] = custom_run_duration
        await self._get("setzone.php", **params)

    async def stop_zone(self, zone: Zone) -> None:
        """Stops a zone.

        :param zone: The zone to stop.
        """
        await self._get("setzone.php", action="stop", relay_id=zone.id)

    async def start_all_zones(
        self,
        controller: Controller,
        mark_run_as_scheduled: bool = False,
        custom_run_duration: int = 0,
    ) -> None:
        """Starts all zones attached to a controller.

        :param controller: The controller whose zones to start.
        :param mark_run_as_scheduled: Whether to mark the zones as having run as scheduled.
        :param custom_run_duration: Duration (in seconds) to run the zones. If not
            specified (or zero), will run for each zone's default configured time.
        """
        _ = mark_run_as_scheduled  # unused
        params = {
            "action": "runall",
            "controller_id": controller.id,
            "period_id": 999,
        }
        if custom_run_duration > 0:
            params["custom"] = custom_run_duration
        await self._get("setzone.php", **params)

    async def stop_all_zones(self, controller: Controller) -> None:
        """Stops all zones attached to a controller.

        :param controller: The controller whose zones to stop.
        """
        await self._get("setzone.php", action="stopall", controller_id=controller.id)

    async def suspend_zone(self, zone: Zone, until: datetime) -> None:
        """Suspends a zone's schedule.

        :param zone: The zone to suspend.
        :param until: When the suspension should end.
        """
        await self._get(
            "setzone.php",
            action="suspend",
            relay_id=zone.id,
            period_id=999,
            custom=int(until.timestamp()),
        )

    async def resume_zone(self, zone: Zone) -> None:
        """Resumes a zone's schedule.

        :param zone: The zone whose schedule to resume.
        """
        await self._get("setzone.php", action="suspend", relay_id=zone.id, period_id=0)

    async def suspend_all_zones(self, controller: Controller, until: datetime) -> None:
        """Suspends the schedule of all zones attached to a given controller.

        :param controller: The controller whose zones to suspend.
        :param until: When the suspension should end.
        """
        await self._get(
            "setzone.php",
            action="suspendall",
            controller_id=controller.id,
            period_id=999,
            custom=int(until.timestamp()),
        )

    async def resume_all_zones(self, controller: Controller) -> None:
        """Resumes the schedule of all zones attached to the given controller.

        :param controller: The controller whose zones to resume.
        """
        await self._get(
            "setzone.php", action="suspendall", period_id=0, controller_id=controller.id
        )

    async def delete_zone_suspension(self, suspension: ZoneSuspension) -> None:
        """Removes a specific zone suspension.

        Useful when there are multiple suspensions for a zone in effect.

        :param suspension: The suspension to delete.
        """
        _ = suspension  # unused
        raise NotImplementedError


def _controller_from_json(controller_json: dict) -> Controller:
    return Controller(
        id=controller_json["controller_id"],
        name=controller_json["name"],
        hardware=ControllerHardware(
            serial_number=controller_json["serial_number"],
        ),
        last_contact_time=datetime.fromtimestamp(controller_json["last_contact"]),
        online=True,
    )


def _zone_from_json(zone_json: dict) -> Zone:
    current_run = None
    next_run = None
    suspended_until = None
    if zone_json["time"] == 1:
        current_run = ScheduledZoneRun(
            remaining_time=timedelta(seconds=zone_json["run"]),
        )
    elif zone_json["time"] == 1576800000:
        suspended_until = datetime.max
    else:
        now = datetime.now().replace(microsecond=0)
        start_time = now + timedelta(seconds=zone_json["time"])
        duration = timedelta(seconds=zone_json["run"])
        next_run = ScheduledZoneRun(
            start_time=start_time,
            end_time=start_time + duration,
            normal_duration=duration,
            duration=duration,
        )
    return Zone(
        id=zone_json["relay_id"],
        number=zone_json["relay"],
        name=zone_json["name"],
        scheduled_runs=ScheduledZoneRuns(
            current_run=current_run,
            next_run=next_run,
        ),
        status=ZoneStatus(
            suspended_until=suspended_until,
        ),
    )


class LegacyHydrawise:
    """Client library for interacting with Hydrawise v1 API.

    This should remain (mostly) compatible with https://github.com/ptcryan/hydrawiser
    """

    def __init__(self, user_token: str, load_on_init: bool = True) -> None:
        self._api_key = user_token
        self.controller_info: dict[str, Any] = {}
        self.controller_status: dict[str, Any] = {}
        if load_on_init:
            self.update_controller_info()

    @property
    def current_controller(self) -> dict:
        controllers = self.controller_info.get("controllers", [])
        if not controllers:
            return {}
        return controllers[0]

    @property
    def status(self) -> str | None:
        return self.current_controller.get("status")

    @property
    def controller_id(self) -> int | None:
        return self.current_controller.get("controller_id")

    @property
    def customer_id(self) -> int | None:
        return self.controller_info.get("customer_id")

    @property
    def num_relays(self) -> int:
        return len(self.controller_status.get("relays", []))

    @property
    def relays(self) -> list[dict]:
        relays = self.controller_status.get("relays", [])
        return sorted(relays, key=lambda r: r["relay"])

    @property
    def relays_by_id(self) -> dict[int, dict]:
        return {r["relay_id"]: r for r in self.controller_status.get("relays", [])}

    @property
    def relays_by_zone_number(self) -> dict[int, dict]:
        return {r["relay"]: r for r in self.controller_status.get("relays", [])}

    @property
    def name(self) -> str | None:
        return self.current_controller.get("name")

    @property
    def sensors(self) -> list[dict]:
        return self.controller_status.get("sensors", [])

    @property
    def running(self) -> str | None:
        return self.controller_status.get("running")

    def update_controller_info(self) -> bool:
        self.controller_info = self._get_controller_info()
        self.controller_status = self._get_controller_status()
        return True

    def _get(self, path: str, **kwargs) -> dict:
        url = f"{_BASE_URL}/{path}"
        params = {"api_key": self._api_key}
        params.update(kwargs)
        resp = requests.get(url, params=params, timeout=_TIMEOUT)

        if resp.status_code != 200:
            resp.raise_for_status()

        resp_json = resp.json()
        if "error_message" in resp_json:
            raise UnknownError(resp_json["error_message"])

        return resp_json

    def _get_controller_info(self) -> dict:
        return self._get("customerdetails.php", type="controllers")

    def _get_controller_status(self) -> dict:
        return self._get("statusschedule.php")

    def suspend_zone(self, days: int, zone: int | None = None) -> dict:
        params: dict[str, Any] = {}

        if days > 0:
            params["custom"] = int(time.time() + (days * 24 * 60 * 60))
            params["period_id"] = 999
        else:
            params["period_id"] = 0

        if zone is None:
            params["action"] = "suspendall"
            return self._get("setzone.php", **params)

        if not self.relays:
            raise NotInitializedError("No zones loaded")

        params["action"] = "suspend"
        params["relay_id"] = self.relays_by_zone_number[zone]["relay_id"]
        return self._get("setzone.php", **params)

    def run_zone(self, minutes: int, zone: int | None = None) -> dict:
        params: dict[str, Any] = {}

        if zone is not None:
            if not self.relays:
                raise NotInitializedError("No zones loaded")
            params["relay_id"] = self.relays_by_zone_number[zone]["relay_id"]
            params["action"] = "run" if minutes > 0 else "stop"
        else:
            params["action"] = "runall" if minutes > 0 else "stopall"

        if minutes > 0:
            params["custom"] = minutes * 60
            params["period_id"] = 999
        else:
            params["period_id"] = 0

        return self._get("setzone.php", **params)
